use std::io;

mod library {
    use serde::{Serialize, Deserialize};
    use std::fs::File;
    use std::io::prelude::*;

    #[derive(Serialize, Deserialize)]
    enum ItemKind {
        Book(String),
        Cd(String),
        Dvd(String),
        Magazine(String),
    }

    #[derive(Serialize, Deserialize)]
    pub struct LoanableItem {
        id: usize,
        title: String,
        return_date: Option<String>,
        kind: ItemKind,
    }

    impl LoanableItem {
        pub fn on_loan(&self) -> bool {
            self.return_date != None
        }

        pub fn return_to_library(&mut self) {
            self.return_date = None
        }

        pub fn set_return_date(&mut self, date: &str) {
            self.return_date = Some(String::from(date));
        }

        fn to_short_string(&self) -> String {
            let item_type = match &self.kind {
                ItemKind::Book(_) => "Book",
                ItemKind::Cd(_) => "CD",
                ItemKind::Dvd(_) => "DVD",
                ItemKind::Magazine(_) => "Magazine",
            };

            let mut result = format!("{}, {}, {}, ", self.id, item_type, &self.title);

            match &self.return_date {
                Some(date) => result.push_str(&format!("yes, {}", date)),
                None => result.push_str(&format!("no")),
            }

            result
        }
    }

    #[derive(Serialize, Deserialize)]
    pub struct Collection {
        items: Vec<LoanableItem>,
    }

    impl Collection {
        pub fn add_new_item(&mut self, kind: char, title: &str, information: &str) {
            let next_id = if self.items.is_empty() {
                0
            } else {
                self.items.iter().max_by_key(|x| x.id).unwrap().id + 1
            };

            match kind {
                'B' => self.create_book(next_id, title, information),
                'C' => self.create_cd(next_id, title, information),
                'D' => self.create_dvd(next_id, title, information),
                'M' => self.create_magazine(next_id, title, information),
                _ => panic!("Program error: unknown kind in add_new_item"),
            };
        }

        fn create_item(&mut self, id: usize, title: &str, kind: ItemKind) {
            self.items.push(LoanableItem {
                id, 
                title: String::from(title),
                return_date: None,
                kind,
            });
        }

        pub fn create_book(&mut self, id: usize, title: &str, author: &str) {
            self.create_item(id, title, ItemKind::Book(String::from(author)));
        }

        pub fn create_cd(&mut self, id: usize, title: &str, artist: &str) {
            self.create_item(id, title, ItemKind::Cd(String::from(artist)));
        }

        pub fn create_dvd(&mut self, id: usize, title: &str, publisher: &str) {
            self.create_item(id, title, ItemKind::Dvd(String::from(publisher)));
        }

        pub fn create_magazine(&mut self, id: usize, title: &str, publisher: &str) {
            self.create_item(id, title, ItemKind::Magazine(String::from(publisher)));
        }

        pub fn find_by_id(&mut self, id: usize) -> Option<&mut LoanableItem> {
            self.items.iter_mut().find(|x| x.id == id)
        }

        pub fn load_collection(&mut self, filename: &str) -> Result<(), std::io::Error> {
            let mut f = File::open(filename)?;
            let mut contents = String::new();
            f.read_to_string(&mut contents)?;
            self.items = serde_json::from_str(&contents)?;
            Ok(())
        }

        pub fn remove_item_by_id(&mut self, id: usize) -> bool {
            for i in 0..self.items.len() {
                if self.items[i].id == id {
                    self.items.remove(i);
                    return true;
                }
            }
            false
        }

        pub fn save_collection(&self, filename: &str) -> Result<(), std::io::Error> {
            let serialised = serde_json::to_string(&self.items)?;
            println!("{}", serialised);
            std::fs::write(filename, serialised)?;
            Ok(())
        }

        fn show_report(&self, name: &str) {
            println!("");
            println!("------------------------------");
            println!("Library Collection Report {}:", name);
            println!("");

            for item in &self.items {
                println!("{}", item.to_short_string());
            }

            println!("");
            println!("TOTAL: {} items with {} on loan", 
                     self.items.len(), 
                     self.items.iter().filter(|i| i.on_loan()).count());
            println!("------------------------------");
        }

        // report lists items in order of increasing ID 
        pub fn show_report_1(&mut self) {
            self.items.sort_by( |a,b| a.id.cmp(&b.id) );
            self.show_report("1");
        }

        // report lists items with on-loan items first, and by date/id after that.
        pub fn show_report_2(&mut self) {
            self.items.sort_by( 
                |a,b|
                match (a.on_loan(), b.on_loan()) {
                    (true, true) => if a.return_date == b.return_date {
                        a.id.cmp(&b.id)
                    } else {
                        a.return_date.cmp(&b.return_date)
                    },
                    (false, false) => a.id.cmp(&b.id),
                    _ => b.on_loan().cmp(&a.on_loan()),
                }
                );
            self.show_report("2");
        }
    }

    pub fn new() -> Collection {
        Collection { items: Vec::new() }
    }
}

fn main() {
    let mut library = sample_library();

    loop {
        show_menu();
        let mut choice = String::new();
        io::stdin().read_line(&mut choice)
            .expect("Missing input");
        match choice.trim().as_ref() {
            "1" => borrow_item(&mut library),
            "2" => return_item(&mut library),
            "3" => library.show_report_1(),
            "4" => library.show_report_2(),
            "5" => add_new_item(&mut library),
            "6" => remove_item(&mut library),
            "7" => save_collection(&library),
            "8" => load_collection(&mut library),
            "q" | "Q" => break,
            _ => println!("Unknown choice selected - try again."),
        }
    }

    println!("Goodbye!");
}

fn add_new_item(library: &mut library::Collection) {
    println!("What kind of item to add? (Book/Cd/Dvd/Magazine)");
    let mut choice = String::new();
    io::stdin().read_line(&mut choice)
        .expect("Missing input for kind of item");
    if choice.len() > 0 {
        let kind = choice.remove(0).to_ascii_uppercase();
        if kind == 'B' || kind == 'C' || kind == 'D' || kind == 'M' {
            println!("Enter item title:");
            let mut title = String::new();
            io::stdin().read_line(&mut title)
                .expect("Missing title for new item");
            println!("Enter item {}:", 
                     match kind {
                         'B' => "author",
                         'C' => "artist",
                         'D' | 'M' => "publisher",
                         _ => "information",
                     });
            let mut information = String::new();
            io::stdin().read_line(&mut information)
                .expect("Missing information for new item");

            library.add_new_item(kind, &title.trim(), &information.trim());
        } else {
            println!(" -- invalid kind entered");
        }
    } else {
        println!(" -- invalid kind entered");
    }
}

fn borrow_item(library: &mut library::Collection) {
    println!("To borrow an item, first find it:");
    let id = enter_find_by_id();
    if let Some(item) = library.find_by_id(id) {  
        if item.on_loan() {
            println!(" -- item {} is already on loan", id);
        } else {
            println!("Enter 'return by' date (YYMMDD):");
            let mut date = String::new();
            io::stdin().read_line(&mut date)
                .expect("Missing input for return by date");
            item.set_return_date(&date.trim());
        }
    } else {
        println!(" -- could not find item {} to borrow", id);
    }
}

fn enter_find_by_id() -> usize {
    println!("Enter ID number of item to return:");
    let mut choice = String::new();
    io::stdin().read_line(&mut choice)
        .expect("Missing input");
    if let Ok(choice) = choice.trim().parse::<usize>() {
        choice
    } else {
        0
    }
}

fn load_collection(library: &mut library::Collection) {
    println!("Name of file to load collection from:");
    let mut filename = String::new();
    io::stdin().read_line(&mut filename)
        .expect("Failed to enter file name to load collection from");
    if let Ok(_) = library.load_collection(&filename.trim()) {
        println!("Loaded the library collection");
    } else {
        println!("Error: could not load the library collection");
    }
}

fn remove_item(library: &mut library::Collection) {
    println!("To remove an item, first find it:");
    let id = enter_find_by_id();
    if !library.remove_item_by_id(id) {
      println!(" -- could not remove item {}", id);
    }
}

fn return_item(library: &mut library::Collection) {
    println!("To return item, first find it:");
    let id = enter_find_by_id();
    if let Some(item) = library.find_by_id(id) {  
        if item.on_loan() {
            item.return_to_library();
            println!(" -- item {} returned", id);
        } else {
            println!(" -- item {} is not on loan!", id);
        }
    } else {
        println!(" -- could not find item {} to return", id);
    }
}

fn save_collection(library: &library::Collection) {
    println!("Name of file to save collection to:");
    let mut filename = String::new();
    io::stdin().read_line(&mut filename)
        .expect("Failed to enter file name to save collection to");
    if let Ok(_) = library.save_collection(&filename.trim()) {
        println!("Saved the library collection");
    } else {
        println!("Error: could not save the library collection");
    }
}

fn show_menu() {
    println!("
Library Management System
=========================

1 Borrow Item
2 Return Item
3 Show Report - id order
4 Show Report - on-loan first

5 Add Item to Collection
6 Remove Item from Collection
7 Save Collection to File
8 Load Collection from File

Q Quit

Enter choice: ");
}

fn sample_library() -> library::Collection {
    let mut library = library::new();
    library.create_book(1, "The Hobbit", "J.R.R. Tolkien");
    library.create_cd(2, "6th Symphony", "G. Mahler");
    library.create_book(3, "Silmarillion", "J.R.R. Tolkien");
    library.create_dvd(4, "Lord of the Rings", "Warner Bros");
    library.create_magazine(5, "The Writers and Readers Magazine", "The Writers Magazine");

    library
}
